"""LBCgo: Data reduction package for the Large Binocular Telescope's Large Binocular Camera (LBC).

Requires astromatic.net software sextractor, scamp, swarp."""
__version__ = '0.1.3'

from .lbcproc import *
from .lbcregister import *
